/*
 * Copyright 2018 Sony Corporation, Sony Video & Sound Products Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nandbank.c
 */

#include <linux/version.h>
#include <linux/mtd/mtd.h>
#include <linux/err.h>


#include "nandbank.h"
#include "nvreg_error.h"

#ifdef DEBUG_NVREG
#define NVREG_PRINT(x...) printk(x)
#else
#define NVREG_PRINT(x...) 
#endif


static char s_empty_entry[NAND_ENTRY_SIZE];
static struct mtd_info *mtd;

static char s_work_page[NAND_PAGE_SIZE];

static unsigned long calc_checksum( char * p_buf, unsigned int len )
{
    int i;
    unsigned long checksum=0;

    //NVREG_PRINT("calc checksum %p (%x)\n", p_buf, len);

    for( i=0; i<len; i++)
    {
        //NVREG_PRINT(" +%x", p_buf[i]);
        checksum += p_buf[i];
    }
    //NVREG_PRINT(" = %lx\n", checksum);
    return checksum;
}


static void dump( char * buf, int len )
{
    int i;
    for( i=0; i<len; i++ )
    {
        if( *(buf+i) == '\0' )
        {
            printk("\\0");
        }
        else
        {
            printk("%c",*(buf+i));
        }
    }
    printk("\n");

}

static void dump_entry( char * buf )
{
    int val_len = *(buf + NAND_ENTRY_SIZE-1 );
    if ( val_len < 64 ) {
	dump( buf, val_len );
    }
}

bool is_same_key( char * p_entry, const char * key )
{
    char * p_separator = NULL;
    unsigned int key_len = 0;
    
    //NVREG_PRINT("key:%s\n", key );
#ifdef DEBUG_NVREG
    dump_entry( p_entry );
#endif
    //char * p_stored_entry_len = p_entry+(NAND_ENTRY_SIZE-1);
	 
    if( strnstr( p_entry, key, NAND_ENTRY_SIZE ) != p_entry )
    {
        NVREG_PRINT("[IsSameKey] key is not match from top\n");
        return false;
    }

    p_separator = strnstr( p_entry, "=", NAND_ENTRY_SIZE );

    if( p_separator == NULL )
    {
        NVREG_PRINT("[IsSameKey] separater =  is not exist\n" );
        return false;
    }

    key_len = p_separator - p_entry;

    if( key_len > NAND_ENTRY_SIZE )
    {
        NVREG_PRINT("[IsSameKey] key_len is invalid (%d)\n", key_len );
        return false;
    }

    if( memcmp( p_entry, key, key_len ) == 0 )
    {
        NVREG_PRINT("[IsSameKey] key is matched\n");
        return true;
    }

    NVREG_PRINT("[IsSameKey] key is not matched\n");
    return false;
}

 

//====================================================================
// PAGE functions
//====================================================================
int page_load( struct ST_PAGE * _page, char * p_page_buf )
{
    size_t read = 0;
    loff_t addr =  _page->addr;
    int ret;
   
    //NVREG_PRINT(" read page[%2d] ( 0x%llx )\n", _page->id, addr );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
    ret = mtd_read( mtd, addr, NAND_PAGE_SIZE, &read, p_page_buf );
#else
    ret = mtd->read( mtd, addr, NAND_PAGE_SIZE, &read, p_page_buf );
#endif
    if( ret != 0 || read != NAND_PAGE_SIZE )
    {
        NVREG_PRINT( "error: read failed at %#llx, ret:%d, read: %zu, pgsize: %d\n",
                (long long)addr , ret, read, NAND_PAGE_SIZE );
    }

    {
        char * checksum_addr = p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_CHECKSUM_OFFSET;
        _page->checksum       = *(unsigned int *)checksum_addr;
    }
    _page->transaction_id = *(unsigned int *)(p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_TRANSACTION_ID_OFFSET);


    return ret;
}

int page_is_match_checksum( struct ST_PAGE * _page, char * p_page_buf )
{
    unsigned int calclated_checksum = calc_checksum( p_page_buf, (NAND_ENTRY_NUM - 1)*NAND_ENTRY_SIZE );

    NVREG_PRINT( "[NVREG] checksum : load:%x, calc:%x\n", _page->checksum, calclated_checksum );
    if( _page->checksum != calclated_checksum )
    {
       return -1;
    }
    return 0;
}

int page_is_empty( struct ST_PAGE * _page, char * p_page_buf )
{
    if( memcmp( p_page_buf, s_empty_entry, NAND_ENTRY_SIZE ) == 0 )
    {
        return 0;
    }
    else
    {
        return -1;
    }
}

int page_add_entry( struct ST_PAGE * _page, char * p_page_buf, const char *key, const void *value, int len, unsigned int transaction_id )
{
    int ret = NVREG_ERR_ENTRY_FULL;
    size_t write = 0;
    int i;
    int entry_len = 0;
    int key_len = strlen( key );
    char buf[NAND_ENTRY_SIZE];
    entry_len = key_len + 1 + len;

    NVREG_PRINT("[NVREG][add] page[%d] try to add %s= ", _page->id, key );
#ifdef DEBUG_NVREG
    dump( (char*)value, len );
#endif
    
    for( i=0; i<NAND_ENTRY_NUM-1; i++ )
    {
        char * entry = p_page_buf + NAND_ENTRY_SIZE * i;

        if( memcmp( entry, s_empty_entry, NAND_ENTRY_SIZE ) == 0 ||
            is_same_key( entry, key ) )
            
        { // hit
            char * p_stored_entry_len = entry+(NAND_ENTRY_SIZE-1);

            NVREG_PRINT("[NVREG][add] entry[%d] add %s\n",i, key );

            NVREG_PRINT("contruct entry\n");
            NVREG_PRINT("key:%s, len:%d\n", key, key_len );
            NVREG_PRINT("val:" );
#ifdef DEBUG_NVREG
            dump( (char*)value, len );
#endif

            memcpy( &buf[0],             key,   key_len );
            memcpy( &buf[key_len],       "=",   1 );
            memcpy( &buf[key_len+1],     value, len );

            NVREG_PRINT("stored_entry_len:%d, entry_len:%d\n", *p_stored_entry_len, entry_len );
         
            if(  *p_stored_entry_len == (char)entry_len  )
            {
                if( memcmp( entry, buf, entry_len ) == 0 )
                {
                    NVREG_PRINT("[NVREG][add] entry[%d] add same key&value. ",i );
#ifdef DEBUG_NVREG
                    dump_entry( entry );
#endif
                    ret = NVREG_OK;
                    break;
                }
            }
            
            *p_stored_entry_len = (char)entry_len;
            memcpy( entry, buf, entry_len );

#ifdef DEBUG_NVREG
            dump_entry( entry );
#endif
            ret = NVREG_OK;
            break;
        }
    }

    if( ret != NVREG_OK )
    {
        return ret;
    }

    {
        _page->checksum = calc_checksum( p_page_buf, (NAND_ENTRY_NUM - 1)*NAND_ENTRY_SIZE );
        *(unsigned int *)(p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_CHECKSUM_OFFSET) = _page->checksum;
        _page->transaction_id = transaction_id;
        *(unsigned int *)(p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_TRANSACTION_ID_OFFSET) = _page->transaction_id;
    }
    NVREG_PRINT( "[NVREG][ADD] checksum:0x%u, transaction_id:%d\n", _page->checksum, transaction_id );

    //NVREG_PRINT("[NVREG][add] mtd->write( addr:%lx, size:0x%x, dist:%p\n", _page->addr, NAND_PAGE_SIZE, p_page_buf );

    // ÍΉFÍLɂƁAnbuffer̂ÍĂ܂̂ŁAWorkobt@gpB
    memcpy( s_work_page, p_page_buf, NAND_PAGE_SIZE );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
    ret = mtd_write(mtd,
                    _page->addr,
                    NAND_PAGE_SIZE,
                    &write,
                    s_work_page );
#else
    ret = mtd->write(mtd,
                     _page->addr,
                     NAND_PAGE_SIZE,
                     &write,
                     s_work_page );
#endif

    //NVREG_PRINT("write 0x%x byte, ret: %d\n", write, ret);

    if (ret != 0 || write != NAND_PAGE_SIZE )
    {
        printk( "[NVREG] error: write failed at page[%d](0x%lx)\n", _page->id, _page->addr );
    }

    return ret;
}

int page_del_entry( struct ST_PAGE * _page, char * p_page_buf, const char *key, unsigned int transaction_id )
{
    int ret = NVREG_ERR_INTERNAL;
    size_t write = 0;
    int i;
    int is_move_from_next_entry = 0;

    NVREG_PRINT("[NVREG][del] page[%d]\n", _page->id );
    
    for( i=0; i<NAND_ENTRY_NUM-1; i++ )
    {
        char * entry = p_page_buf + NAND_ENTRY_SIZE * i;

        if( is_move_from_next_entry == 0 &&
            is_same_key( entry, key ) )
        { // hit
            NVREG_PRINT("[NVREG][del] entry[%d].key = %s\n", i, key );
            is_move_from_next_entry = 1;
            ret = NVREG_OK;
        }

        if( is_move_from_next_entry == 1 )
        {
            if( i == (NAND_ENTRY_LAST_ID -1 ) )
            {
                memset( entry, 0xFF, NAND_ENTRY_SIZE );
            }
            else
            {
                memcpy( entry, entry+NAND_ENTRY_SIZE, NAND_ENTRY_SIZE );
            }
        }
    }

    if( ret != NVREG_OK )
    {
        NVREG_PRINT("[NVREG][del] no hit key = %s\n", key );
        return NVREG_ERR_NO_ENTRY;
    }

    {
        _page->checksum = calc_checksum( p_page_buf, (NAND_ENTRY_NUM - 1)*NAND_ENTRY_SIZE );        
        *(unsigned int *)(p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_CHECKSUM_OFFSET) = _page->checksum;
        _page->transaction_id = transaction_id;
        *(unsigned int *)(p_page_buf + (NAND_ENTRY_LAST_ID * NAND_ENTRY_SIZE ) + NAND_TRANSACTION_ID_OFFSET) = _page->transaction_id;
    }
    NVREG_PRINT("[NVREG][del] checksum:0x%u, transaction_id:%d\n", _page->checksum, transaction_id );

    //NVREG_PRINT("[NVREG][del] mtd->write( addr:%lx, size:0x%x, dist:%p\n", _page->addr, NAND_PAGE_SIZE, p_page_buf );

    // ÍΉFÍLɂƁAnbuffer̂ÍĂ܂̂ŁAWorkobt@gpB
    memcpy( s_work_page, p_page_buf, NAND_PAGE_SIZE );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
    ret = mtd_write(mtd,
                    _page->addr,
                    NAND_PAGE_SIZE,
                    &write,
                    s_work_page );
#else
    ret = mtd->write(mtd,
                     _page->addr,
                     NAND_PAGE_SIZE,
                     &write,
                     s_work_page );
#endif

    //NVREG_PRINT("write 0x%x byte, ret: %d\n", write, ret);

    if (ret != 0 || write != NAND_PAGE_SIZE )
    {
        printk( "[NVREG] error: write failed at page[%d](0x%lx)\n", _page->id, _page->addr );
    }

    return ret;

}

void page_init( struct ST_PAGE * _page, unsigned long base_addr, int page_id )
{
    _page->id = page_id;
    _page->addr = base_addr + NAND_PAGE_SIZE * page_id;

    _page->load               = page_load;
    _page->is_match_checksum  = page_is_match_checksum;
    _page->is_empty           = page_is_empty;
    _page->add_entry          = page_add_entry;
    _page->del_entry          = page_del_entry;
}


//====================================================================
// BLOCK functions
//====================================================================
void block_erase( struct ST_BLOCK * _block )
{
	int err;
	struct erase_info ei;

	memset(&ei, 0, sizeof(struct erase_info));
	ei.mtd  = mtd;
	ei.addr = _block->addr;
	ei.len  = NAND_BLOCK_SIZE;


    NVREG_PRINT("[NVREG] MTD : erase %llx (size : %llx)\n", ei.addr, ei.len );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	err = mtd_erase( mtd, &ei);
#else
	err = mtd->erase( mtd, &ei);
#endif
	if (err)
    {
	    printk( "[NVREG] Erasing Block is failed !!! %d \n", err );
	    return;
    }

	if (ei.state == MTD_ERASE_FAILED)
    {
	    printk( "some erase error[%d] occurred at erasing block.\n", ei.state );
	}
}

int block_load_active_page( struct ST_BLOCK * _block, struct ST_PAGE * p_active_page, char * p_page_buf )
{
    int page_id = 0;
    int ret = 0;

    NVREG_PRINT("[NVREG] load block[%d]\n", _block->id );
    
    for( page_id=NAND_PAGE_NUM-1; page_id>=0; page_id-- )
    {  // 1pageƓǂݍ
        struct ST_PAGE * page = &_block->page[page_id];

        page->load( page, p_page_buf );

        if( page->is_empty( page, p_page_buf ) == 0 )
        {
            if( page_id == 0 )
            {
                NVREG_PRINT("[NVREG] it seems fresh BLOCK\n");
                p_active_page->id = 0;
                p_active_page->transaction_id = 0;
                _block->read_page_id  = 0;
                _block->write_page_id = 0;
                ret = -1;
                break;
            }
            else
            {
                continue;
            }
        }
        else
        {
            NVREG_PRINT("[NVREG] hit!!! valid page[%d] \n", page_id);
            if( page->is_match_checksum( page, p_page_buf ) < 0 )
            { // skip broken page
                NVREG_PRINT( "[WARNING] checksums are not match !!! => skip page[%d]\n", page_id);
                continue;
            }

            p_active_page->id = page_id;
            p_active_page->transaction_id = page->transaction_id;
            _block->read_page_id  = page_id;
            _block->write_page_id = page_id+1;
            if( page_id == (NAND_PAGE_NUM -1) )
            {
                _block->write_page_id = 0;
                _block->is_page_full = 1;
            }
            ret = 0;
            break;

        }
    }

    NVREG_PRINT( "[NVREG][load page] read_page_id:%d, check_sum:0x%x, transaction_id:%d, write_page_id:%d, is_need_clear:%s\n",
                 _block->read_page_id,
                 p_active_page->checksum,
                 p_active_page->transaction_id,
                 _block->write_page_id,
                 _block->is_page_full == 1 ? "o" : "x" );

    return ret;
}


int block_add_entry( struct ST_BLOCK * _block, char * p_page_buf, const char *key, const void *value, int len, unsigned int transaction_id )
{
    int ret = NVREG_OK;
    struct ST_PAGE *page;

    NVREG_PRINT("[NVREG][add] block[%d]\n", _block->id );

    if( _block->is_page_full == 1 )
    {
        _block->erase( _block );
        _block->is_page_full = 0;
        _block->write_page_id = 0;
    }

    page = &_block->page[_block->write_page_id];
    ret = page->add_entry( page, p_page_buf, key, value, len, transaction_id );
    if( ret != NVREG_OK )
    {
        return ret;
    }

    _block->read_page_id = _block->write_page_id;
    _block->write_page_id ++;


    if( _block->write_page_id >= NAND_PAGE_NUM )
    {
        _block->is_page_full = 1;
    }
    
    return ret;
}

int block_del_entry( struct ST_BLOCK * _block, char * p_page_buf, const char *key, unsigned int transaction_id )
{
    int ret = NVREG_OK;
    struct ST_PAGE *page;

    NVREG_PRINT("[NVREG][del] block[%d]\n", _block->id );

    if( _block->is_page_full == 1 )
    {
        _block->erase( _block );
        _block->is_page_full = 0;
        _block->write_page_id = 0;
    }

    page = &_block->page[_block->write_page_id];
    ret = page->del_entry( page, p_page_buf, key, transaction_id );
    if( ret != NVREG_OK )
    {
        return ret;
    }
    
    _block->read_page_id = _block->write_page_id;
    _block->write_page_id ++;

    if( _block->write_page_id >= NAND_PAGE_NUM )
    {
        _block->is_page_full = 1;
    }
    
    return ret;

}


void block_init( struct ST_BLOCK * _block, int block_id )
{
    int i=0;

    _block->id = block_id;
    _block->addr = NAND_BLOCK_SIZE * block_id;

    _block->erase            = block_erase;
    _block->load_active_page = block_load_active_page;
    _block->add_entry        = block_add_entry;
    _block->del_entry        = block_del_entry;

    for( i=0; i<NAND_PAGE_NUM; i++)
    {
        struct ST_PAGE * page = &_block->page[i];
        page->init = page_init;
        page->init( page, _block->addr, i );
    }

    _block->is_page_full = 0;
    _block->read_page_id = 0;
    _block->write_page_id = 0;

}



//====================================================================
// BANK functions
//====================================================================
void bank_load_active_page( struct ST_BANK * _bank )
{
    struct ST_BLOCK * block=NULL;
    bool ret[2];
    struct ST_PAGE page[2];

    NVREG_PRINT("[NVREG] ==========================================\n");
    NVREG_PRINT("[NVREG] load bank[%d]\n", _bank->id );

    ret[0] = _bank->block[0].load_active_page( &_bank->block[0], &page[0], &_bank->page_buf[0][0] ); 
    ret[1] = _bank->block[1].load_active_page( &_bank->block[1], &page[1], &_bank->page_buf[0][0] ); 

    if( ret[0] < 0 && ret[1] < 0 )
    {
        _bank->active_block_id = 0;
        memset( &_bank->page_buf[0][0], 0xFF, NAND_PAGE_SIZE );
        return;        
    }

    if( ret[0] == 0 && ret[1] == 0 )
    {
        NVREG_PRINT("[NVREG] bank[%d] - block[%d]:o block[%d]:o\n",
               _bank->id,
               _bank->block[0].id,
               _bank->block[1].id );
        NVREG_PRINT("[NVREG] bank[%d] - block[%d].transaction_id = %d, block[%d].transaction_id = %d\n",
               _bank->id,
               _bank->block[0].id, page[0].transaction_id,
               _bank->block[1].id, page[1].transaction_id );

        if( page[0].transaction_id >= page[1].transaction_id )
        {
            _bank->active_block_id = 0;
        }
        else
        {
            _bank->active_block_id = 1;
        }
    }
    else
    {
        if( ret[0] == 0 )
        {
            NVREG_PRINT("[NVREG] bank[%d] - block[%d]:o block[%d]:x\n",
                        _bank->id,
                        _bank->block[0].id,
                        _bank->block[1].id );
            _bank->active_block_id = 0;
        }
        if( ret[1] == 0)
        {
            NVREG_PRINT("[NVREG] bank[%d] - block[%d]:x block[%d]:o\n",
                        _bank->id,
                        _bank->block[0].id,
                        _bank->block[1].id );
            _bank->active_block_id = 1;
        }
    }

    NVREG_PRINT("[NVREG] reload active bank[%d]\n", _bank->id );
    block = &_bank->block[_bank->active_block_id];
    block->load_active_page( block, &page[_bank->active_block_id], &_bank->page_buf[0][0] );

    return;
}


void bank_dump_active_page( struct ST_BANK * _bank )
{
    int i;

#ifdef DEBUG_NVREG
    int j;
    char line[256];
    char * checksum_addr = &_bank->page_buf[NAND_ENTRY_LAST_ID][NAND_CHECKSUM_OFFSET];
    unsigned int loaded_checksum = *(unsigned int *)(checksum_addr );
#endif

    NVREG_PRINT( "dump active block[%d].page[%d] - checksum: 0x%x\n",
            _bank->block[_bank->active_block_id].id,
            _bank->block[_bank->active_block_id].read_page_id,
            loaded_checksum );

    for( i=0; i<NAND_ENTRY_NUM; i++)
    {
#ifdef DEBUG_NVREG
        char * p_line = line;
        p_line += snprintf( p_line, 128, "[%2d] %06x: ", i, i*NAND_ENTRY_SIZE);

        for( j=0; j<NAND_ENTRY_SIZE; j++ )
        {
            p_line += snprintf( p_line, 128, "%02x", _bank->page_buf[i][j] );
        }
        NVREG_PRINT( "entry%s :", line );
#endif
        if( memcmp( &_bank->page_buf[i][0], s_empty_entry, NAND_ENTRY_SIZE ) != 0 &&
            i != ( NAND_ENTRY_NUM -1 ) )
        {
            dump_entry( &_bank->page_buf[i][0] );
        }
        else
        {
            NVREG_PRINT( "\n");
            //return;
        }
    }
    return;
}



int bank_add_entry( struct ST_BANK * _bank, const char *key, const void *value, int len)
{
    unsigned int transaction_id;
    struct ST_BLOCK * block = &_bank->block[ _bank->active_block_id];
    int active_block_id = _bank->active_block_id;
    int ret;

    NVREG_PRINT("[NVREG][add] block is_page_full:%d\n", block->is_page_full );

    transaction_id = block->page[block->read_page_id].transaction_id + 1;

    NVREG_PRINT("[NVREG][add] bank:%d active_block_id:%d transaction_id:%d\n", _bank->id , active_block_id, transaction_id );

    if( block->is_page_full == 1 )
    {
        if( _bank->active_block_id == 0 )
        {
            active_block_id = 1;
        }
        else
        {
            active_block_id = 0;
        }
        block = &_bank->block[ active_block_id ];
    }

    if( ( ret = block->add_entry( block, &_bank->page_buf[0][0], key, value, len, transaction_id ) ) == NVREG_OK ){
	_bank->active_block_id = active_block_id;
    }
    return ret;
}

int bank_get_entry( struct ST_BANK * _bank, const char *key, void *buf, int maxlen)
{
    int i;
    char * p_entry     = NULL;
    char * p_separator = NULL;
    int entry_len=0, key_len=0, val_len=0;

    NVREG_PRINT( "[NVREG][get] bank[%d] block[%d]\n", _bank->id, _bank->active_block_id );
    NVREG_PRINT( "[NVREG][get] key       : %s\n", key );
    NVREG_PRINT( "[NVREG][get] maxlen    : %d\n", maxlen );

    for( i=0; i<NAND_ENTRY_NUM-1; i++ )
    {
        p_entry = &_bank->page_buf[i][0];
        if( strnstr( p_entry, key, NAND_ENTRY_SIZE ) != NULL  )
        {
            if( (p_separator = strnstr( p_entry, "=", NAND_ENTRY_SIZE )) != NULL  ) 
            {
                NVREG_PRINT("[NVREG][get] entry no  : %d\n", i );
                NVREG_PRINT("[NVREG][get] entry     : " );
#ifdef DEBUG_NVREG
                dump_entry( p_entry );
#endif
                NVREG_PRINT("[NVREG][get] maxlen    : %d\n",maxlen);
                entry_len = (int)(*(p_entry + (NAND_ENTRY_SIZE-1)));
                NVREG_PRINT("[NVREG][get] entry_len : %d\n", entry_len);

                key_len = p_separator - p_entry;
                NVREG_PRINT("[NVREG][get] key_len   : %d\n", key_len);

                val_len = entry_len - key_len -1;
                NVREG_PRINT("[NVREG][get] val_len   : %d\n", val_len);

                if( val_len < 0 ||  val_len > maxlen )
                {
                    printk("[NVREG][get] ERROR : val_len:%d, max_len:%d\n", val_len, maxlen );
                    return NVREG_ERR_INVALID_PARAM;
                }

                memcpy( buf, (p_separator+1), val_len );
                return val_len;
            }
        }
    }
    return NVREG_ERR_NO_ENTRY;
}

int bank_del_entry( struct ST_BANK * _bank, const char *key )
{
    unsigned int transaction_id;
    struct ST_BLOCK * block = &_bank->block[ _bank->active_block_id];
    int active_block_id = _bank->active_block_id;
    int ret;

    NVREG_PRINT("[NVREG][del] block is_page_full:%d\n", block->is_page_full );

    transaction_id = block->page[block->read_page_id].transaction_id + 1;

    NVREG_PRINT("[NVREG][del] bank:%d active_block_id:%d transaction_id:%d\n", _bank->id , active_block_id, transaction_id );

    if( block->is_page_full == 1 )
    {
        if( _bank->active_block_id == 0 )
        {
            active_block_id = 1;
        }
        else
        {
            active_block_id = 0;
        }
        block = &_bank->block[ active_block_id ];
    }

    if( ( ret = block->del_entry( block, &_bank->page_buf[0][0], key, transaction_id ) ) == NVREG_OK ){
	_bank->active_block_id = active_block_id;
    }
    return ret;
}




void bank_init( struct ST_BANK * _bank, int bank_id )
{
    _bank->load_active_page = bank_load_active_page;
    _bank->dump_active_page = bank_dump_active_page;
    _bank->add_entry        = bank_add_entry;
    _bank->get_entry        = bank_get_entry;
    _bank->del_entry        = bank_del_entry;

    _bank->block[0].init = block_init;
    _bank->block[1].init = block_init;

    _bank->block[0].init( &_bank->block[0], bank_id*2 );
    _bank->block[1].init( &_bank->block[1], bank_id*2+1 );

    _bank->id = bank_id;
    _bank->active_block_id = 0;

    memset( _bank->page_buf, 0xFF, (NAND_ENTRY_NUM*NAND_ENTRY_SIZE) );

}



//====================================================================
// global functions
//====================================================================
int nandbank_init( void )
{
    int mtd_no = 20;
    int err;

    memset( s_empty_entry, 0xFF, NAND_ENTRY_SIZE );

    printk( "use /dev/mtd%d\n", mtd_no );

    mtd = get_mtd_device(NULL, mtd_no );
    if (IS_ERR(mtd)) {
        err = PTR_ERR(mtd);
        printk("error: Cannot get MTD device\n");
        return NVREG_ERR_INTERNAL;
    }

    printk( "total size  : %llu\n", mtd->size );
    printk( "block size  : %d\n",   mtd->erasesize );
    printk( "page  size  : %d\n",   mtd->writesize );


    if( mtd->size < (NAND_BLOCK_SIZE * NAND_BLOCK_COUNT) ||
        mtd->writesize != NAND_PAGE_SIZE ||
        mtd->erasesize != NAND_BLOCK_SIZE )
    {
        printk("unexpected nand device !!!\n");
    }

    printk( "block count : %llu\n", mtd->size      / mtd->erasesize );
    printk( "page count  : %d\n",   mtd->erasesize / mtd->writesize );

    return NVREG_OK;
}

