/*
 * Copyright 2018 Sony Corporation, Sony Video & Sound Products Inc.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * nandreg.c
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/semaphore.h>

#include "nvreg_error.h"

#ifdef DEBUG_NVREG
#define NVREG_PRINT(x...) printk(x)
#else
#define NVREG_PRINT(x...) 
#endif

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include "nandbank.h"

#define NAND_BANK_NUM ( 4 )

static struct semaphore s_nandreg_sem;
static struct ST_BANK s_bank[NAND_BANK_NUM];

int nandreg_init( void )
{
    sema_init( &s_nandreg_sem, 1 );
    return 0;
}


int nandreg_load( void )
{
    int i = 0;

    down( &s_nandreg_sem );
    nandbank_init();

    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        s_bank[i].init = bank_init;
        s_bank[i].init( &s_bank[i], i );
    }

    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        s_bank[i].load_active_page( &s_bank[i] );
#ifdef DEBUG_NVREG
        s_bank[i].dump_active_page( &s_bank[i] );
#endif
    }

    up( &s_nandreg_sem );
    return NVREG_OK;

}

int nandreg_add(const char *key, const void *value, int len)
{
    int key_len;
    int i = 0;
    int ret = NVREG_ERR_INTERNAL;
    
    NVREG_PRINT("[NVREG] ADD ===============\n");

    if( len < 0 )
    {
        printk("[NVREG][add] ERROR : length of value is %d\n", len );
        return NVREG_ERR_INVALID_PARAM;
    }
    
    key_len = strlen( key );
    if( (key_len+len+2/*'='&'\n'*/) > NAND_ENTRY_SIZE )
    {
        printk("[NVREG][add] ERROR : key_len:%d, val_len: %d\n", key_len, len );
        return NVREG_ERR_INVALID_PARAM;
    }

    down( &s_nandreg_sem );
    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        if( (ret = s_bank[i].add_entry( &s_bank[i], key, value, len )) == NVREG_OK )
        {
            break;
        }
    }
    up( &s_nandreg_sem );
    return ret;
}
EXPORT_SYMBOL( nandreg_add );

int nandreg_get(const char *key, void *buf, int maxlen)
{
    int i = 0; 
    int ret = NVREG_ERR_INTERNAL;

    NVREG_PRINT("[NVREG] GET ===============\n");
    
    down( &s_nandreg_sem );
    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        if( (ret = s_bank[i].get_entry( &s_bank[i], key, buf, maxlen )) != NVREG_ERR_NO_ENTRY )
        {
            break;
        }
    }
    up( &s_nandreg_sem );
    return ret;
}
EXPORT_SYMBOL( nandreg_get );

int nandreg_dump( void )
{
    int i = 0; 
    NVREG_PRINT("[NVREG] DUMP ===============\n");

    down( &s_nandreg_sem );
    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        s_bank[i].dump_active_page( &s_bank[i] );
    }
    up( &s_nandreg_sem );
    return NVREG_OK;
}


int nandreg_del(const char *key)
{
    int ret = NVREG_ERR_INTERNAL;
    int i = 0; 

    NVREG_PRINT("[NVREG] DEL ===============\n");

    down( &s_nandreg_sem );
    for( i=0; i < NAND_BANK_NUM; i++ )
    {
        if( (ret = s_bank[i].del_entry( &s_bank[i], key )) == NVREG_OK )
        {
            break;
        }
    }
    up( &s_nandreg_sem );
    return ret;
}
EXPORT_SYMBOL( nandreg_del );
